from setuptools import setup

setup(name='spikesorting_tsne',
      version='1.0.0',
      description='T-sne with burnes hut and cuda extension (with python wrappers and python code for spike sorting)',
      long_description='This is the repository for the python part of the conda python package that allows running the results of automated spike sorting algorithms through the t-SNE algorithm obtaining a 2D or 3D embedding of the spikes. \
						Although this package offers some functionality dedicated to spikesorting the t-SNE part of it is kept separate and can be run with any matrix of samples x features.\
						The package is split into two parts. The python part (in this repo) has the following functionality:\
						1. It has the main function through which the whole algorithm is called.\
						2. The main function wraps the C++ executable (Barnes_Hut.exe) and calls it appropriately.\
						3. It uses numba gpu to create the spike (or any other sample matrix) distances.\
						4. It offers functions that operate on the result of the kilosort spikesorting algorithm to create the required samples (spikes) x features (template distances) matrix for t-SNE to run on.\
						5. It offers functions that allow the user to split a dataset of spikes that is too lrage for the algorithm to sub groups that can be seperately run and then recombined (uses information generated by ythe kilosort algorithm)\
						The C++ part (which generates the Barnes_Hut.exe executable) can be found [here](https://github.com/georgedimitriadis/spikesorting_tsne_bhpart).\
						More detailed documentation can be found in the [Github Pages](https://georgedimitriadis.github.io/spikesorting_tsne/) of this repo.',
      long_description_content_type='text/markdown',
      url='https://github.com/georgedimitriadis/spikesorting_tsne',
      author='George Dimitriadis',
      author_email='gdimitri@hotmail.com',
      packages=['spikesorting_tsne'],
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Visualization',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    project_urls={
        'Documentation': 'https://georgedimitriadis.github.io/spikesorting_tsne/',
        'Python code repository': 'https://github.com/georgedimitriadis/spikesorting_tsne',
        'C++ code repository': 'https://github.com/georgedimitriadis/spikesorting_tsne_bhpart',
    },
    python_requires='>=3.5',
    zip_safe=False)


